@@ -38,13 +38,18 @@
  */
 
 #if defined WIN32 || defined __WIN32
+# undef HAVE_PROGRESS
 # undef __STRICT_ANSI__
 # include <sys/types.h>
 # include <sys/utime.h>
 #else
+# define HAVE_PROGRESS
 # ifndef _POSIX_SOURCE
 #  define _POSIX_SOURCE
 # endif
+# ifndef _POSIX_C_SOURCE
+#  define _POSIX_C_SOURCE 200112L /* for popen(3) */
+# endif
 # include <sys/types.h>
 # include <utime.h>
 #endif
@@ -2193,6 +2198,12 @@ main (int argc, char **argv)
   char* imagedir = 0;
   /** directory for Exif thumbnail files (0=output disabled) */
   char* thumbdir = 0;
+#ifdef HAVE_PROGRESS
+  /** progress indicator */
+  FILE* progress = 0;
+  /** progress counter */
+  int progresscnt = 0;
+#endif /* HAVE_PROGRESS */
 
   srcinfo.err = jpeg_std_error (&jsrcerr.pub);
   jsrcerr.pub.addon_message_table = exif_message_table;
@@ -2218,6 +2229,18 @@ main (int argc, char **argv)
       continue;
     }
 
+#ifdef HAVE_PROGRESS
+    if (progress) {
+      if (0 > fprintf (progress, "%u\n",
+		       100 * (progresscnt - argc) / progresscnt) ||
+	  fflush (progress)) {
+	perror ("write(progress)");
+	status = 2;
+	goto done;
+      }
+    }
+#endif /* HAVE_PROGRESS */
+
     while (opts && *++opts) { /* process all flags */
       switch (*opts) {
       default:
@@ -2229,7 +2252,7 @@ main (int argc, char **argv)
       case '?':
       case 'h':
 	(void) fputs
-	  ("exifiron 1.2.3\n"
+	  ("exifiron 1.2.4\n"
 	   "Orientation correction and shrinking of Exif files\n"
 	   "Usage: exifiron [options] file ...\n"
 	   "Options:\n"
@@ -2271,7 +2294,13 @@ main (int argc, char **argv)
 	   "-s dir\n"
 	   "\tsave the image files in dir instead of replacing originals\n"
 	   "-S dir\n"
-	   "\tsave the Exif thumbnails in dir\n", stderr);
+	   "\tsave the Exif thumbnails in dir\n"
+#ifdef HAVE_PROGRESS
+	   "-P command\n"
+	   "\toutput progress indication to an external command\n"
+#endif /* HAVE_PROGRESS */
+	   "--\tdisable processing further options\n",
+	   stderr);
 	goto done;
 
       case '0': transformation = 0; break;
@@ -2297,6 +2326,32 @@ main (int argc, char **argv)
 
       case 'n': optimize = 0; break;
 
+#ifdef HAVE_PROGRESS
+      case 'P':
+	if (argc <= 2) {
+	  putc ('-', stderr), putc (*opts, stderr);
+	  (void) fputs (": missing argument (use -h for help)\n", stderr);
+	  status = 1;
+	  goto done;
+	}
+	if (progress)
+	  pclose (progress);
+	argc--;
+	fflush (stderr);
+	fflush (stdout);
+	if (*++argv) {
+	  progress = popen (*argv, "w");
+	  if (!progress) {
+	    fprintf (stderr, "-P %s", *argv);
+	    perror (": popen");
+	  }
+	}
+	else
+	  progress = 0;
+	progresscnt = argc;
+	break;
+#endif /* HAVE_PROGRESS */
+
       case 's': case 'S':
 	if (argc <= 2) {
 	  putc ('-', stderr), putc (*opts, stderr);
@@ -2419,5 +2474,9 @@ main (int argc, char **argv)
  done:
   jpeg_destroy_compress (&dstinfo);
   jpeg_destroy_decompress (&srcinfo);
+#ifdef HAVE_PROGRESS
+  if (progress)
+    pclose (progress);
+#endif /* HAVE_PROGRESS */
   return status;
 }
