@@ -3,7 +3,7 @@
  * @author Marko Mkel <marko.makela@iki.fi>
  */
 
-/* Copyright  2003,2004,2005 Marko Mkel.
+/* Copyright  2003,2004,2005,2006 Marko Mkel.
 
    This file is part of PHOTOMOLO, a program for generating
    thumbnail images and HTML files for browsing digital photographs.
@@ -341,6 +341,7 @@ html_parent_links (FILE* file,
 /** Generate links to child directories
  * @param file	the output stream
  * @param tree	the directory tree
+ * @param pflen	number of first bytes to ignore in tree->name
  * @param ixnam	index.html file name
  * @param spec	the thumbnail directory name
  * @param blen	base directory name length
@@ -349,6 +350,7 @@ html_parent_links (FILE* file,
 static int
 html_child_links (FILE* file,
 		  const struct dirtree* tree,
+		  unsigned pflen,
 		  const char* ixnam,
 		  const char* spec,
 		  unsigned blen)
@@ -359,13 +361,14 @@ html_child_links (FILE* file,
   if (html_raw (file, "<dl>\n"))
     return 1;
   for (i = 0; i < tree->n_dirs; i++) {
-    const char* base = strrchr (tree->dirs[i].name, '/');
+    const struct dirtree* dir = tree->dirs + i;
+    const char* base = strrchr (dir->name + pflen, '/');
     if (base)
       base++;
     else
-      base = tree->dirs[i].name;
+      base = dir->name + pflen;
     if (html_raw (file, "<dt><a href=\"../") ||
-	html_uri (file, tree->dirs[i].name + blen + 1) ||
+	html_uri (file, dir->name + pflen + blen + 1) ||
 	html_raw (file, "/") ||
 	html_uri (file, spec) ||
 	html_raw (file, "/") ||
@@ -374,19 +377,36 @@ html_child_links (FILE* file,
 	html_esc (file, base, 0) ||
 	html_raw (file, "</a>"))
       return 1;
-    if (tree->dirs[i].n_files || tree->dirs[i].n_others) {
-      if (tree->dirs[i].n_others)
-	fprintf (file, " (%u+%u)",
-		 tree->dirs[i].n_files,
-		 tree->dirs[i].n_others);
-      else
-	fprintf (file, " (%u)", tree->dirs[i].n_files);
+    if (dir->n_files || dir->n_files_dirs) {
+      unsigned n_others =
+	dir->n_files - dir->n_jpeg - dir->n_images;
+      unsigned n_others_dirs =
+	dir->n_files_dirs - dir->n_jpeg_dirs - dir->n_images_dirs;
+      /* display the numbers of files in current directory */
+      if ((0 > fprintf (file, " (%u", dir->n_jpeg)) ||
+	  ((dir->n_images || n_others) &&
+	   0 > fprintf (file, "+%u", dir->n_images)) ||
+	  (n_others &&
+	   0 > fprintf (file, "+%u", n_others)))
+	return 1;
+      /* display the numbers of files in subdirectories */
+      if (dir->n_files_dirs) {
+	if ((0 > fprintf (file, "; <em>%u", dir->n_jpeg_dirs)) ||
+	    ((dir->n_images_dirs || n_others_dirs) &&
+	     0 > fprintf (file, "+%u", dir->n_images_dirs)) ||
+	    (n_others_dirs &&
+	     0 > fprintf (file, "+%u", n_others_dirs)) ||
+	    fputs ("</em>", file) == EOF)
+	  return 1;
+      }
+      if (fputs (")", file) == EOF)
+	return 1;
     }
     if (html_raw (file, "</dt>\n"))
       return 1;
-    if (tree->dirs[i].n_dirs) {
+    if (dir->n_dirs) {
       if (html_raw (file, "<dd>") ||
-	  html_child_links (file, &tree->dirs[i], ixnam, spec, blen) ||
+	  html_child_links (file, dir, pflen, ixnam, spec, blen) ||
 	  html_raw (file, "</dd>\n"))
 	return 1;
     }
@@ -399,6 +419,7 @@ html_child_links (FILE* file,
  * @param argc	number of command-line arguments
  * @param argv	command-line arguments
  * @param tree	the directory
+ * @param pflen	number of first bytes to ignore in tree->name
  * @param level	number of '/' characters in the root directory name
  * @param info	image dimensions
  * @param comm	image comments
@@ -415,6 +436,7 @@ dumphtml (FILE* file,
 	  int argc,
 	  char** argv,
 	  const struct dirtree* tree,
+	  unsigned pflen,
 	  unsigned level,
 	  const struct imageinfo* info,
 	  const struct comment*const* comm,
@@ -433,15 +455,18 @@ dumphtml (FILE* file,
   /** file modification time */
   char mtime[20];
   /** length of file name */
-  unsigned fnlen = strlen (tree->name) + 1;
+  unsigned fnlen = strlen (tree->name + pflen) + 1;
   /** length of the thumbnail directory name */
   unsigned speclen = strlen (spec->thumbdir) + 1;
-  const char* title = strrchr (tree->name, '/');
+  const char* title = strrchr (tree->name + pflen, '/');
+  /** flag: listed any files yet? */
+  int listed;
+
   if (title)
     title++;
   else
-    title = tree->name;
-  for (rootname = tree->name; level && *rootname; rootname++)
+    title = tree->name + pflen;
+  for (rootname = tree->name + pflen; level && *rootname; rootname++)
     if (*rootname == '/')
       level--;
   if (html_raw (file,
@@ -489,13 +514,10 @@ dumphtml (FILE* file,
 		"<body>\n<h1>") ||
       html_parent_links (file, rootname, opts->html, spec->thumbdir) ||
       html_esc (file, title, 0) ||
-      html_raw (file, "</h1>\n"
-		"<table><tr>\n"
-		"<td>") ||
-      html_child_links (file, tree, opts->html, spec->thumbdir,
-			strlen (tree->name)) ||
-      html_raw (file, "</td>\n"
-		"<td><p>") ||
+      html_raw (file, "</h1>\n") ||
+      html_child_links (file, tree, pflen, opts->html, spec->thumbdir,
+			strlen (tree->name + pflen)) ||
+      html_raw (file, "<p>") ||
       html_a_specs (file, opts->html, thumb, spec) ||
       html_raw (file, "</p>\n"))
     return 1;
@@ -507,92 +529,119 @@ dumphtml (FILE* file,
     return 1;
   }
   /* prepare the thumbnail file name */
-  memcpy (fn, tree->name, fnlen - 1);
+  memcpy (fn, tree->name + pflen, fnlen - 1);
   memcpy (fn + fnlen, spec->thumbdir, speclen - 1);
   fn[fnlen - 1] = fn[fnlen + speclen - 1] = '/';
 
-  /* write the image descriptions */
-  if (tree->n_files &&
-      html_raw (file, "<p class=\"img\">")) {
-  error:
-    free (fn);
-    return 1;
-  }
+  /* list the image files */
+  listed = 0;
+
   for (i = 0; i < tree->n_files; i++) {
     /** image dimensions */
     char dim[20];
+    /** thumbnail image information */
+    struct imageinfo thumb_info;
+    /** thumbnail file name (converted or original) */
+    const char* name;
+    /** thumbnail file */
+    FILE* f;
+    /** thumbnail file name length */
+    unsigned namelen;
+    char* s;
+    const struct fileinfo* const fi = tree->files + i;
+
+    switch (fi->kind) {
+    case FILE_JPEG:
+    case FILE_IMAGE:
+      if (!info[i].height || !info[i].width)
+	continue; /* skip invalid file */
+      break;
+    case FILE_OTHER:
+      continue;
+    }
+
+    if (!listed) {
+      listed = 1;
+      if (html_raw (file, "<p class=\"img\">"))
+	goto error;
+    }
 
-    if (!info[i].height || !info[i].width)
-      continue; /* skip invalid file */
-    sprintf (dim, "%u&times;%u", info[i].width, info[i].height);
     if (opts->notim ||
 	!strftime (mtime, sizeof mtime, "%Y-%m-%d %H:%M:%S",
-		   localtime (&tree->files[i].mtime)))
+		   localtime (&fi->mtime)))
       *mtime = 0;
-    if (html_raw (file, "<a href=\"../") ||
-	html_uri (file, tree->files[i].name) ||
+
+    name = fi->converted ? fi->converted : fi->name;
+    sprintf (dim, "%u&times;%u", info[i].width, info[i].height);
+    if (html_raw (file, "<a\nhref=\"../") ||
+	html_uri (file, fi->name) ||
 	html_raw (file, "\"><img src=\"") ||
-	html_uri (file, tree->files[i].name) ||
+	html_uri (file, name) ||
 	html_raw (file, "\"\nalt=\"[") ||
 	html_info (file, mtime, dim, comm[i]) ||
 	html_raw (file, "]\"\ntitle=\"") ||
 	html_info (file, mtime, dim, comm[i]) ||
 	html_raw (file, "\""))
       goto error;
-    else {
-      /** thumbnail image information */
-      struct imageinfo thumb_info;
-      /** thumbnail file */
-      FILE* f;
-      /** thumbnail file name length */
-      unsigned namelen = strlen (tree->files[i].name) + 1;
-      char* s = realloc (fn, fnlen + speclen + namelen);
-      if (!s)
-	goto memory;
-      fn = s;
-      memcpy (fn + fnlen + speclen, tree->files[i].name, namelen);
-      if (!(f = fopen (fn, "rb"))) {
-	(void) fputs (fn, stderr), (void) fflush (stderr);
-	perror (": open (reading)");
-	goto error;
-      }
-      s = (char*) readinfo (&thumb_info, 0, f);
-      (void) fclose (f);
-      if (s) {
-	(void) fputs (fn, stderr);
-	(void) fputs (": ", stderr), (void) fputs (s, stderr);
-	(void) putc ('\n', stderr);
-      }
-      else
-	(void) fprintf (file, "\nwidth=\"%u\" height=\"%u\"",
-			thumb_info.width, thumb_info.height);
+    namelen = strlen (name) + 1;
+    s = realloc (fn, fnlen + speclen + namelen);
+    if (!s)
+      goto memory;
+    fn = s;
+    memcpy (fn + fnlen + speclen, name, namelen);
+    if (!(f = fopen (fn, "rb"))) {
+      (void) fputs (fn, stderr), (void) fflush (stderr);
+      perror (": open (reading)");
+      goto error;
     }
-
-    if (html_raw (file, "></a>\n"))
+    s = (char*) readinfo (&thumb_info, 0, f);
+    (void) fclose (f);
+    if (s) {
+      (void) fputs (fn, stderr);
+      (void) fputs (": ", stderr), (void) fputs (s, stderr);
+      (void) putc ('\n', stderr);
+    }
+    else
+      (void) fprintf (file, "\nwidth=\"%u\" height=\"%u\"",
+		      thumb_info.width, thumb_info.height);
+    if (html_raw (file, "></a>"))
       goto error;
   }
-  if (tree->n_files && html_raw (file, "</p>\n"))
-    goto error;
 
-  /* list other files */
-  for (i = 0; i < tree->n_others; i++) {
+  if (listed && html_raw (file, "</p>\n")) {
+  error:
+    free (fn);
+    return 1;
+  }
+
+  /* list the non-image files */
+
+  for (i = 0; i < tree->n_files; i++) {
+    const struct fileinfo* const fi = tree->files + i;
+
+    switch (fi->kind) {
+    case FILE_JPEG:
+    case FILE_IMAGE:
+      continue; /* skip image files */
+    case FILE_OTHER:
+      break;
+    }
+
     if (opts->notim ||
 	!strftime (mtime, sizeof mtime, "%Y-%m-%d %H:%M:%S",
-		   localtime (&tree->others[i].mtime)))
+		   localtime (&fi->mtime)))
       *mtime = 0;
+
     if (html_raw (file, "<p><a href=\"../") ||
-	html_uri (file, tree->others[i].name) ||
+	html_uri (file, fi->name) ||
 	(*mtime && (html_raw (file, "\"\ntitle=\"") ||
 		    html_raw (file, mtime))) ||
 	html_raw (file, "\">") ||
-	html_esc (file, tree->others[i].name, 0) ||
-	html_raw (file, "</a></p>\n")) {
-      free (fn);
-      return 1;
-    }
+	html_esc (file, fi->name, 0) ||
+	html_raw (file, "</a></p>\n"))
+      goto error;
   }
+
   free (fn);
-  return html_raw (file, "</td>\n"
-		   "</tr></table>"
-		   "</body></html>\n");
+  return html_raw (file, "</body></html>\n");
 }
