@@ -25,13 +25,18 @@
    59 Temple Place, Suite 330, Boston, MA 02111 USA. */
 
 #if defined WIN32 || defined __WIN32
+# undef HAVE_PROGRESS
 # undef __STRICT_ANSI__
 # include <sys/types.h>
 # include <sys/utime.h>
 #else
+# define HAVE_PROGRESS
 # ifndef _POSIX_SOURCE
 #  define _POSIX_SOURCE
 # endif
+# ifndef _POSIX_C_SOURCE
+#  define _POSIX_C_SOURCE 200112L /* for popen(3) */
+# endif
 # include <sys/types.h>
 # include <utime.h>
 #endif
@@ -759,6 +764,12 @@ main (int argc, char **argv)
   char* dstname = 0;
   /** thumbnail file name */
   char* thumbname = 0;
+#ifdef HAVE_PROGRESS
+  /** progress indicator */
+  FILE* progress = 0;
+  /** progress counter */
+  int progresscnt = 0;
+#endif /* HAVE_PROGRESS */
 
   srcinfo.err = jpeg_std_error (&jsrcerr.pub);
   jsrcerr.pub.addon_message_table = exif_message_table;
@@ -783,6 +794,18 @@ main (int argc, char **argv)
       continue;
     }
 
+#ifdef HAVE_PROGRESS
+    if (progress) {
+      if (0 > fprintf (progress, "%u\n",
+		       100 * (progresscnt - argc) / progresscnt) ||
+	  fflush (progress)) {
+	perror ("write(progress)");
+	status = 2;
+	goto done;
+      }
+    }
+#endif /* HAVE_PROGRESS */
+
     while (opts && *++opts) { /* process all flags */
       switch (*opts) {
       default:
@@ -794,7 +817,7 @@ main (int argc, char **argv)
       case '?':
       case 'h':
 	(void) fputs
-	  ("jpegnail 1.2.3\n"
+	  ("jpegnail 1.2.4\n"
 	   "Remove or embed thumbnails in Exif or JFIF JPEG images\n"
 	   "Usage: jpegnail [options] file ...\n"
 	   "Options:\n"
@@ -809,13 +832,20 @@ main (int argc, char **argv)
 	   "-b"
 	   "\tbaseline JPEG\n"
 	   "-p"
-	   "\tprogressive JPEG (default)\n"
-	   "-n"
+	   "\tprogressive JPEG (default)\n",
+	   stderr);
+	(void) fputs
+	  ("-n"
 	   "\tdo not compute optimal Huffman coding for the image\n"
 	   "-S dir\n"
 	   "\tread image thumbnails from dir instead of computing them\n"
 	   "-s dir\n"
-	   "\tsave the image files in dir instead of replacing originals\n",
+	   "\tsave the image files in dir instead of replacing originals\n"
+#ifdef HAVE_PROGRESS
+	   "-P command\n"
+	   "\toutput progress indication to an external command\n"
+#endif /* HAVE_PROGRESS */
+	   "--\tdisable processing further options\n",
 	   stderr);
 	goto done;
 
@@ -824,8 +854,29 @@ main (int argc, char **argv)
 
       case 'n': optimize = 0; break;
 
-      case 's':
-      case 'S':
+#ifdef HAVE_PROGRESS
+      case 'P':
+	if (argc <= 2)
+	  goto missing;
+	if (progress)
+	  pclose (progress);
+	argc--;
+	fflush (stderr);
+	fflush (stdout);
+	if (*++argv) {
+	  progress = popen (*argv, "w");
+	  if (!progress) {
+	    fprintf (stderr, "-P %s", *argv);
+	    perror (": popen");
+	  }
+	}
+	else
+	  progress = 0;
+	progresscnt = argc;
+	break;
+#endif /* HAVE_PROGRESS */
+
+      case 's': case 'S':
 	if (argc <= 2) {
 	missing:
 	  (void) putc ('-', stderr), (void) putc (*opts, stderr);
@@ -980,5 +1031,9 @@ main (int argc, char **argv)
  done:
   jpeg_destroy_compress (&dstinfo);
   jpeg_destroy_decompress (&srcinfo);
+#ifdef HAVE_PROGRESS
+  if (progress)
+    pclose (progress);
+#endif /* HAVE_PROGRESS */
   return status;
 }
