@@ -4,7 +4,7 @@
  * @author Marko Mkel <marko.makela@iki.fi>
  */
 
-/* Copyright  2003,2004,2005 Marko Mkel.
+/* Copyright  2003,2004,2005,2006 Marko Mkel.
 
    This file is part of PHOTOMOLO, a program for generating
    thumbnail images and HTML files for browsing digital photographs.
@@ -28,6 +28,19 @@
  * photomolo -T _8/s8 -T _4/s4 -T _120/h120 -T _240/h240 -T _480/h480 cd1
  */
 
+#if defined WIN32 || defined __WIN32
+# undef HAVE_PROGRESS
+#else
+# define HAVE_PROGRESS
+# ifndef _POSIX_C_SOURCE
+#  define _POSIX_C_SOURCE 200112L /* for popen(3) */
+# endif
+# ifndef _BSD_SOURCE
+#  define _BSD_SOURCE /* for strdup(3) */
+# endif
+#endif
+#define HAVE_CONVERTER
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -37,8 +50,15 @@
 #include "info.h"
 #include "spec.h"
 #include "html.h"
+#ifdef HAVE_CONVERTER
+# include <sys/stat.h> /* open(2) */
+# include <sys/types.h>
+# include <fcntl.h>
+# include <unistd.h> /* close(2), unlink(2) */
+# include "setenv.h"
+#endif /* HAVE_CONVERTER */
 
-#define PHOTOMOLO_VERSION "pHoToMoLo 1.2.3"
+#define PHOTOMOLO_VERSION "pHoToMoLo 1.2.4"
 
 /** The libjpeg decompression context */
 static struct jpeg_decompress_struct cin;
@@ -59,7 +79,7 @@ static struct html_opts hopts = {
   0, /* "shortcut icon" file name */
   "iso-8859-1", /* character set */
   PHOTOMOLO_VERSION ";\n "
-  "http://www.funet.fi/pub/sci/graphics/packages/photomolo/photomolo.html;\n"
+  "http://www.iki.fi/~msmakela/software/photomolo/photomolo.html;\n"
   /* contents of the HTML META Generator element */
 };
 
@@ -78,6 +98,31 @@ static int main_argc;
 /** command-line arguments */
 static char** main_argv;
 
+#ifdef HAVE_LINK
+/** Path to linked image file tree, or NULL */
+static const char* linkdir = 0;
+/** strlen (linkdir), or 0 */
+static unsigned linkdirlen = 0;
+/** Flag: nonzero=generate symlinks to linkdir, zero=generate hard links */
+static int linkdir_sym = 0;
+#else
+# define linkdirlen 0
+#endif /* HAVE_LINK */
+
+#ifdef HAVE_PROGRESS
+/** progress indicator */
+static FILE* progress;
+/** number of image files processed */
+static unsigned progresscnt;
+/** number of image files to process */
+static unsigned progresstotal;
+#endif /* HAVE_PROGRESS */
+
+#ifdef HAVE_CONVERTER
+/** name of external JPEG converter program */
+static const char* converter;
+#endif /* HAVE_CONVERTER */
+
 /** Cleanup function */
 static void
 cleanup (void)
@@ -92,29 +137,43 @@ cleanup (void)
   /* deallocate the libjpeg contexts */
   jpeg_destroy_compress (&cout);
   jpeg_destroy_decompress (&cin);
+#ifdef HAVE_PROGRESS
+  /* close the progress indicator */
+  if (progress)
+    pclose (progress);
+#endif /* HAVE_PROGRESS */
 }
 
 /** Create the thumbnail images
- * @param dir	the directory
- * @param info	the image information (for pruning non-image files)
+ * @param dir	the directory; dir->n_images may be updated
+ * @param info	(in) the image information (for pruning non-JPEG files),
+ *		(out) image information from converted non-JPEG files
+ * @param comm	(out) image comment from converted non-JPEG files
  * @return	zero on success, nonzero on failure
  */
 static int
-make_thumbnails (const struct dirtree* dir,
-		 const struct imageinfo* info)
+make_thumbnails (struct dirtree* dir,
+		 struct imageinfo* info,
+		 struct comment** comm)
 {
   unsigned i;
   char* fn = 0;
-  unsigned fnlen = strlen (dir->name) + 1;
+  unsigned fnlen = strlen (dir->name + linkdirlen) + 1;
+#ifdef HAVE_CONVERTER
+  char* jpgfn = 0;
+#endif /* HAVE_CONVERTER */
 
   if (!(fn = malloc (fnlen + 1))) {
   memory:
     free (fn);
+#ifdef HAVE_CONVERTER
+    free (jpgfn);
+#endif /* HAVE_CONVERTER */
     (void) fputs ("out of memory\n", stderr);
     return 1;
   }
 
-  memcpy (fn, dir->name, fnlen);
+  memcpy (fn, dir->name + linkdirlen, fnlen - 1);
   fn[fnlen - 1] = '/';
 
   for (i = 0; i < dir->n_files; i++) {
@@ -122,20 +181,161 @@ make_thumbnails (const struct dirtree* dir,
     FILE* input;
     char* s;
     unsigned namelen;
-    if (!info[i].width || !info[i].height)
-      continue; /* no image */
-    namelen = strlen (dir->files[i].name) + 1;
+    struct fileinfo* file = dir->files + i;
+#ifdef HAVE_PROGRESS
+    progresscnt++;
+    if (progress && progresstotal) {
+      if (0 > fprintf (progress, "%u\n",
+		       100 * progresscnt / progresstotal) ||
+	  fflush (progress)) {
+	perror ("write (progress)");
+	free (fn);
+# ifdef HAVE_CONVERTER
+	free (jpgfn);
+# endif /* HAVE_CONVERTER */
+	return 4;
+      }
+    }
+#endif /* HAVE_PROGRESS */
+    if (file->kind == FILE_OTHER)
+      continue;
+    namelen = strlen (file->name) + 1;
     if (!(s = realloc (fn, fnlen + namelen)))
       goto memory;
     fn = s;
-    memcpy (fn + fnlen, dir->files[i].name, namelen);
-    if (!(input = fopen (fn, "rb"))) {
-      (void) fputs (fn, stderr), (void) fflush (stderr);
+    memcpy (fn + fnlen, file->name, namelen);
+#ifdef HAVE_CONVERTER
+    if (file->kind == FILE_IMAGE && converter) {
+      /* Convert the file to JPEG.  First convert the file name. */
+      unsigned basenamelen;
+      int fd;
+      const char suffix[] = ".jpg";
+# if 0 /* Just append the suffix to better avoid collisions. */
+      const char* end = strrchr (file->name, '.');
+      if (end) {
+	/* Try to replace the last dot-separated component of
+	the file name with the .jpg suffix. */
+
+	basenamelen = end - file->name;
+	if (!basenamelen /* There is no dot-separated component in the name */
+	    || !strcmp (end, suffix) /* The name already ends in .jpg */)
+	  goto append_to_name;
+      }
+      else
+      append_to_name:
+# endif
+	/* Append the suffix to the original file name. */
+	basenamelen = namelen - 1; /* account for the terminating NUL */
+
+      /* Generate the file name for the conversion */
+      if (!(s = realloc (jpgfn, fnlen + basenamelen + sizeof suffix)))
+	goto memory;
+      jpgfn = s;
+      memcpy (jpgfn, fn, fnlen + basenamelen);
+      memcpy (jpgfn + fnlen + basenamelen, suffix, sizeof suffix);
+
+      free (file->converted);
+      if (!(file->converted = strdup (jpgfn + fnlen)))
+	goto memory;
+
+#if 0 /* This would not produce HTML thumbnails on subsequent runs. */
+      /* Determine whether the thumbnail images have
+      already been generated for this image, in order to
+      avoid invoking the external converter. */
+      {
+	unsigned jpgnamelen = strlen (file->converted) + 1;
+	char* tfn = malloc (fnlen);
+	char* t;
+	if (!tfn)
+	  goto memory;
+	memcpy (tfn, fn, fnlen);
+
+	for (spec = thumbs; spec; spec = spec->next) {
+	  unsigned speclen = strlen (spec->thumbdir) + 1;
+	  if (!(t = realloc (tfn, fnlen + jpgnamelen + speclen))) {
+	    free (tfn);
+	    goto memory;
+	  }
+	  tfn = t;
+	  memcpy (tfn + fnlen, spec->thumbdir, speclen);
+	  tfn[fnlen + speclen - 1] = '/';
+	  memcpy (tfn + fnlen + speclen, file->converted, jpgnamelen);
+	  if (!exists_mtime (tfn, file->mtime)) {
+	    free (t);
+	    goto do_convert;
+	  }
+	}
+
+	free (tfn);
+      }
+      /* No need to convert */
+      continue;
+
+    do_convert:
+#endif
+      /* Try creating the file exclusively */
+      fd = open (jpgfn, O_WRONLY | O_CREAT | O_EXCL, 0666);
+      if (fd < 0) {
+	(void) fputs (jpgfn, stderr), (void) fflush (stderr);
+	perror (": open (creating)");
+	free (file->converted);
+	file->converted = 0;
+	continue;
+      }
+      if (close (fd)) {
+	(void) fputs (jpgfn, stderr), (void) fflush (stderr);
+	perror (": close (creating)");
+	free (file->converted);
+	file->converted = 0;
+	continue;
+      }
+      if (set_environment ("IMG_IN", fn) ||
+	  set_environment ("IMG_OUT", jpgfn) ||
+	  system (converter)) {
+	free (file->converted);
+	file->converted = 0;
+	file->kind = FILE_OTHER;
+	goto remove_converted;
+      }
+    }
+#endif /* HAVE_CONVERTER */
+    if (!(input = fopen (s, "rb"))) {
+      (void) fputs (s, stderr), (void) fflush (stderr);
       perror (": open");
       free (fn);
+#ifdef HAVE_CONVERTER
+      free (jpgfn);
+#endif /* HAVE_CONVERTER */
       return 2;
     }
-    jpeg_stdio_src (&cin, input);
+#ifdef HAVE_CONVERTER
+    if (s == jpgfn) {
+      if (set_environment ("IMG_IN", 0) ||
+	  set_environment ("IMG_OUT", 0))
+	goto close_input;
+      s = (char*) readinfo (info + i, comm + i, input);
+      if (s) {
+	/* error while reading image info */
+	(void) fputs (jpgfn + fnlen, stderr);
+	(void) fputs (": ", stderr), (void) fputs (s, stderr);
+	(void) putc ('\n', stderr);
+	freeinfo (comm[i]); comm[i] = 0;
+	/* clear the info structure for this file */
+	memset (info + i, 0, sizeof *info);
+	/* ignore this file */
+	(void) fclose (input);
+	free (file->converted);
+	file->converted = 0;
+	file->kind = FILE_OTHER;
+	if (unlink (jpgfn))
+	  goto remove_failed;
+	continue;
+      }
+      rewind (input);
+      namelen = strlen (file->converted) + 1;
+      dir->n_images++;
+    }
+#endif /* HAVE_CONVERTER */
     for (spec = thumbs; spec; spec = spec->next) {
       unsigned speclen = strlen (spec->thumbdir) + 1;
       if (!(s = realloc (fn, fnlen + namelen + speclen))) {
@@ -148,22 +348,30 @@ make_thumbnails (const struct dirtree* dir,
       errexit:
 	(void) fclose (input);
 	free (fn);
+#ifdef HAVE_CONVERTER
+	free (jpgfn);
+#endif /* HAVE_CONVERTER */
 	return 3;
       }
       fn[fnlen + speclen - 1] = '/';
-      memcpy (fn + fnlen + speclen, dir->files[i].name, namelen);
-      if (!exists_mtime (fn, dir->files[i].mtime)) {
+      memcpy (fn + fnlen + speclen, file->converted
+	      ? file->converted
+	      : file->name,
+	      namelen);
+      if (!exists_mtime (fn, file->mtime)) {
 	FILE* output = fopen (fn, "wb");
 	if (!output) {
 	  (void) fputs (fn, stderr), (void) fflush (stderr);
 	  perror (": open (writing)");
 	  goto errexit;
 	}
+	rewind (input);
+	jpeg_stdio_src (&cin, input);
 	jpeg_stdio_dest (&cout, output);
 	cout.image_width = spec->max_width;
 	cout.image_height = spec->max_height;
 	if (scale_image (&cin, &cout, spec->shrink, quality) ||
-	    touch_mtime (fn, dir->files[i].mtime)) {
+	    touch_mtime (fn, file->mtime)) {
 	  (void) fclose (output);
 	  goto errexit;
 	}
@@ -172,22 +380,80 @@ make_thumbnails (const struct dirtree* dir,
 	  perror (": close");
 	  goto errexit;
 	}
-	if (spec->next)
-	  rewind (input);
       }
     }
+#ifdef HAVE_CONVERTER
+  close_input:
+    s = jpgfn && *jpgfn ? jpgfn : fn;
     if (fclose (input)) {
-      memcpy (fn + fnlen, dir->files[i].name, namelen);
+      (void) fputs (s, stderr), (void) fflush (stderr);
+      perror (": close"); /* report but otherwise ignore the error */
+    }
+  remove_converted:
+    if (jpgfn && *jpgfn && unlink (jpgfn)) {
+    remove_failed:
+      (void) fputs (jpgfn, stderr), (void) fflush (stderr);
+      perror (": unlink");
+      free (fn);
+      free (jpgfn);
+      return 3;
+    }
+    if (jpgfn)
+      *jpgfn = 0;
+#else /* HAVE_CONVERTER */
+    if (fclose (input)) {
+      memcpy (fn + fnlen, file->name, namelen);
       (void) fputs (fn, stderr), (void) fflush (stderr);
       perror (": close"); /* report but otherwise ignore the error */
     }
+#endif /* HAVE_CONVERTER */
   }
 
   free (fn);
+#ifdef HAVE_CONVERTER
+  free (jpgfn);
+#endif /* HAVE_CONVERTER */
   return 0;
 }
 
-/** Dump the HTML code (or remove thumbnail directories if remove_thumbdirs)
+/** Remove the thumbnail directories
+ * @param tree	the directory
+ * @return	zero on success, nonzero on failure
+ */
+static int
+remove_thumbs (const struct dirtree* tree)
+{
+  int status = 0;
+  const struct thumb_spec* spec;
+  unsigned dirnamelen = strlen (tree->name + linkdirlen) + 1;
+
+  char* filename = malloc (dirnamelen + 1);
+  if (!filename) {
+  memory:
+    (void) fputs ("out of memory\n", stderr);
+    return 1;
+  }
+
+  memcpy (filename, tree->name + linkdirlen, dirnamelen - 1);
+  filename[dirnamelen - 1] = '/';
+
+  for (spec = thumbs; spec; spec = spec->next) {
+    unsigned speclen = strlen (spec->thumbdir) + 1;
+    char* fn = realloc (filename, dirnamelen + speclen + indexnamelen + 1);
+    if (!fn)
+      goto memory;
+    filename = fn;
+    memcpy (filename + dirnamelen, spec->thumbdir, speclen);
+    if (remove_directory (filename)) {
+      status = 2;
+      break;
+    }
+  }
+  free (filename);
+  return status;
+}
+
+/** Dump the HTML code
  * @param tree	the directory
  * @param level	number of '/' characters in the root directory name
  * @param info	image dimensions
@@ -208,7 +474,7 @@ dumphtmls (const struct dirtree* tree,
 {
   int status = 0;
   const struct thumb_spec* spec;
-  unsigned dirnamelen = strlen (tree->name) + 1;
+  unsigned dirnamelen = strlen (tree->name + linkdirlen) + 1;
 
   char* filename = malloc (dirnamelen + 1);
   if (!filename) {
@@ -216,7 +482,8 @@ dumphtmls (const struct dirtree* tree,
     (void) fputs ("out of memory\n", stderr);
     return 1;
   }
-  memcpy (filename, tree->name, dirnamelen - 1);
+
+  memcpy (filename, tree->name + linkdirlen, dirnamelen - 1);
   filename[dirnamelen - 1] = '/';
 
   for (spec = thumbs; spec; spec = spec->next) {
@@ -227,27 +494,20 @@ dumphtmls (const struct dirtree* tree,
       goto memory;
     filename = fn;
     memcpy (filename + dirnamelen, spec->thumbdir, speclen);
-    if (remove_thumbdirs) {
-      if (remove_directory (filename)) {
-	free (filename);
-	return 2;
-      }
-      continue;
-    }
-    if (make_directory (filename)) {
-      free (filename);
-      return 2;
-    }
+    if (make_directory (filename))
+      goto errexit;
     filename[dirnamelen + speclen - 1] = '/';
     memcpy (filename + dirnamelen + speclen, hopts.html, indexnamelen);
     if (!(f = fopen (filename, "wb"))) {
       (void) fputs (filename, stderr), (void) fflush (stderr);
       perror (": open");
-      free (filename);
-      return 1;
+    errexit:
+      status = 2;
+      break;
     }
     status = dumphtml (f, main_argc, main_argv,
-		       tree, level, info, comm, &hopts, thumbs, spec,
+		       tree, linkdirlen,
+		       level, info, comm, &hopts, thumbs, spec,
 		       top, prev, next);
     if (fclose (f) || status) {
       (void) fputs (filename, stderr), (void) fflush (stderr);
@@ -262,7 +522,7 @@ dumphtmls (const struct dirtree* tree,
 }
 
 /** Dump the image file directory listing
- * @param tree	the directory
+ * @param tree	the directory; n_images may be updated
  * @param level	number of '/' characters in the root directory name
  * @param top	name of parent directory
  * @param prev	name of previous directory
@@ -270,7 +530,7 @@ dumphtmls (const struct dirtree* tree,
  * @return	zero on success, nonzero on failure
  */
 static int
-dumpfiles (const struct dirtree* tree,
+dumpfiles (struct dirtree* tree,
 	   unsigned level,
 	   const char* top,
 	   const char* prev,
@@ -286,17 +546,59 @@ dumpfiles (const struct dirtree* tree,
   struct comment** comment;
   /** directory name length */
   unsigned dirnamelen;
-  /** file name for reading tree->files */
+  /** total length of "../" components in dirnamelen */
+  unsigned uplen = 0;
+#ifdef HAVE_LINK
+  /** path component leading to parent directory */
+  static const char up[3] = "../";
+#endif /* HAVE_LINK */
+  /** path name for reading tree->files and creating links */
   char* filename;
 
   dirnamelen = strlen (tree->name);
+
+#ifdef HAVE_LINK
+  /* creating relative symlinks? */
+  if (linkdirlen && linkdir_sym && *tree->name != '/') {
+    /* then prefix the path with "../" components */
+    const char* s;
+    unsigned level = 1;
+    for (s = tree->name + linkdirlen; *s; s++)
+      if (*s == '/')
+	level++;
+    dirnamelen += uplen = level * sizeof up;
+  }
+#endif /* HAVE_LINK */
+
   filename = malloc (dirnamelen + 2);
   if (!filename) {
   memory:
     (void) fputs ("out of memory\n", stderr);
     return 1;
   }
-  memcpy (filename, tree->name, dirnamelen);
+
+#ifdef HAVE_LINK
+  if (uplen) {
+    /* Prefix the file name with "../" components */
+    unsigned i;
+    for (i = 0; i < uplen; i += sizeof up)
+      memcpy (filename + i, up, sizeof up);
+  }
+#endif /* HAVE_LINK */
+
+  memcpy (filename + uplen, tree->name, dirnamelen - uplen);
+
+#ifdef HAVE_LINK
+  if (linkdirlen) {
+    /* Create the base directory in the output tree. */
+    filename[dirnamelen] = 0;
+    if (make_directory (filename + uplen + linkdirlen)) {
+      free (filename);
+      return 2;
+    }
+  }
+#endif /* HAVE_LINK */
+
   filename[dirnamelen++] = '/';
 
   /* gather image info */
@@ -322,8 +624,19 @@ dumpfiles (const struct dirtree* tree,
       }
       filename = s;
       memcpy (filename + dirnamelen, tree->files[i].name, len);
-      if (!(f = fopen (filename, "rb"))) {
-	(void) fputs (filename, stderr), (void) fflush (stderr);
+
+#ifdef HAVE_LINK
+      if (linkdirlen &&
+	  create_link (filename, filename + uplen + linkdirlen, linkdir_sym))
+	continue; /* ignore unlinkable files */
+#endif /* HAVE_LINK */
+
+      if (tree->files[i].kind != FILE_JPEG)
+	continue; /* ignore non-JPEG files here */
+
+      if (!(f = fopen (filename + uplen + linkdirlen, "rb"))) {
+	(void) fputs (filename + uplen + linkdirlen, stderr);
+	(void) fflush (stderr);
 	perror (": fopen");
 	continue; /* ignore unreadable files */
       }
@@ -331,7 +644,7 @@ dumpfiles (const struct dirtree* tree,
       (void) fclose (f);
       if (s) {
 	/* ignore errors while reading image info */
-	(void) fputs (filename, stderr);
+	(void) fputs (filename + uplen + linkdirlen, stderr);
 	(void) fputs (": ", stderr), (void) fputs (s, stderr);
 	(void) putc ('\n', stderr);
 	freeinfo (comment[i]); comment[i] = 0;
@@ -342,14 +655,19 @@ dumpfiles (const struct dirtree* tree,
   }
   free (filename);
 
-  /* generate the thumbnails */
-  if (!status && !remove_thumbdirs)
-    status = make_thumbnails (tree, info);
-  /* generate the HTML or remove the thumbnail directories */
-  if (!status && (remove_thumbdirs || dump_html))
-    status = dumphtmls (tree, level,
-			info, (const struct comment**) comment,
-			top, prev, next);
+  if (remove_thumbdirs) {
+    if (!status)
+      status = remove_thumbs (tree);
+  }
+  else {
+    /* generate the thumbnails */
+    if (!status)
+      status = make_thumbnails (tree, info, comment);
+    if (!status && dump_html)
+      status = dumphtmls (tree, level,
+			  info, (const struct comment**) comment,
+			  top, prev, next);
+  }
 
   /* deallocate the image comments */
   for (i = tree->n_files; i--; )
@@ -360,7 +678,7 @@ dumpfiles (const struct dirtree* tree,
 }
 
 /** Dump the image file directory tree
- * @param tree	the directory tree
+ * @param tree	the directory tree; n_images and n_images_dirs may be updated
  * @param level	number of '/' characters in the root directory name
  * @param top	name of parent directory
  * @param prev	name of previous directory
@@ -368,20 +686,27 @@ dumpfiles (const struct dirtree* tree,
  * @return	zero on success, nonzero on failure
  */
 static int
-dumptree (const struct dirtree* tree,
+dumptree (struct dirtree* tree,
 	  unsigned level,
 	  const char* top,
 	  const char* prev,
 	  const char* next)
 {
   unsigned i;
-  for (i = 0; i < tree->n_dirs; i++)
+  for (i = 0; i < tree->n_dirs; i++) {
+#ifdef HAVE_LINK
+    if (linkdirlen && make_directory (tree->name + linkdirlen))
+      return 2;
+#endif /* HAVE_LINK */
     if (dumptree (tree->dirs + i,
 		  level,
 		  tree->name,
 		  i > 0 ? tree->dirs[i - 1].name : 0,
 		  i < tree->n_dirs - 1 ? tree->dirs[i + 1].name : 0))
       return 1;
+    tree->n_images_dirs +=
+      tree->dirs[i].n_images_dirs + tree->dirs[i].n_images;
+  }
   return dumpfiles (tree, level, top, prev, next);
 }
 
@@ -399,6 +724,39 @@ prune_dirs (const char* name)
   return 0;
 }
 
+/** identify files
+ * @param name	file name
+ * @return	type of the file
+ */
+static enum file_type
+identify_images (const char* name)
+{
+  FILE* f;
+  if (!(f = fopen (name, "rb"))) {
+    (void) fputs (name, stderr), (void) fflush (stderr);
+    perror (": open (reading)");
+    return FILE_OTHER;
+  }
+  else {
+    int is_jpeg =
+      getc (f) == 0xff &&
+      getc (f) == 0xd8;
+    int is_eof = feof (f);
+    (void) fclose (f);
+    if (is_jpeg)
+      return FILE_JPEG;
+    if (is_eof)
+      return FILE_OTHER;
+  }
+
+#ifdef HAVE_CONVERTER
+  /* Assume that all files can be converted into an image. */
+  if (converter)
+    return FILE_IMAGE;
+#endif /* HAVE_CONVERTER */
+  return FILE_OTHER;
+}
+
 /** create the CSS file if it does not exist
  * @param path	the path name to the directory
  * @return	0 if successful
@@ -410,10 +768,11 @@ create_css (const char* path)
   static const char css1[] =
     "/*\n"
     " * Sample style sheet for pHoToMoLo-generated image gallery HTML\n"
-    " * Written in 2003 by Marko Mkel (marko.makela@iki.fi).\n"
+    " * Written in 2003,2006 by Marko Makela (marko.makela (at) iki.fi).\n"
     " * This file is in the public domain.\n"
     " * The layout has been tested with the following browsers:\n"
-    " * Lynx, Netscape 4, Opera 6, IE 5.01, IE 5.5, IE 6.0, Mozilla 1.x.\n"
+    " * Lynx, Opera 6, Opera 8.5, IE 5.01, IE 5.5, IE 6.0, Mozilla 1.x.\n"
+    " * This layout does not work well on Netscape 4.\n"
     " */\n"
     "\n"
     "a:link,a:visited,a:active,a:hover{color:red;background:black;}\n"
@@ -425,6 +784,7 @@ create_css (const char* path)
     "body{color:white;background:black;"
     "font-family:verdana,helvetica,sans-serif;}\n"
     "body,h1,p,table,tr,td,dl,img{border:0;margin:0;}\n"
+    "p.img{line-height:0;}\n"
     "\n"
     "h1{text-align:left;font-weight:bold;font-size:150%;}\n"
     "\n"
@@ -432,9 +792,7 @@ create_css (const char* path)
     "dl{font-weight:bold;font-style:normal;}\n"
     "dl dl{font-weight:normal;}\n"
     "dl dl dl{font-style:oblique;}\n"
-    "dl dl dl dl{font-size:smaller;}\n"
-    "\n"
-    "tr,td{vertical-align:top;margin:0;border:0;}\n";
+    "dl dl dl dl{font-size:smaller;}\n";
   FILE* f;
   unsigned pathlen;
   char* s;
@@ -499,15 +857,15 @@ main (int argc, char** argv)
        "-H htmlfile\n"
        "\tname of HTML files (default=\"index.html\")\n"
        "-C cssfile\n"
-       "\tname of CSS files (default=\"index.css\")\n",
-       stderr);
-    (void) fputs
-      ("-c charset\n"
+       "\tname of CSS files (default=\"index.css\")\n"
+       "-c charset\n"
        "\tHTML and JPEG comment character set (default=\"iso-8859-1\")\n"
        "-i iconfile\n"
        "-I file.ico\n"
-       "\tname of icon files (default=none)\n"
-       "-t"
+       "\tname of icon files (default=none)\n",
+       stderr);
+    (void) fputs
+      ("-t"
        "\tgenerate thumbnails only (no HTML or CSS)\n"
        "-r"
        "\tremove the thumbnail directories listed with -T\n"
@@ -515,6 +873,20 @@ main (int argc, char** argv)
        "\tJPEG compression quality (default=75)\n"
        "-d"
        "\tomit timestamps from the HTML\n"
+#ifdef HAVE_LINK
+       "-l ../path\n"
+       "\tgenerate symlinks to the image files\n"
+       "-L ../path\n"
+       "\tgenerate hard links to the image files\n"
+#endif /* HAVE_LINK */
+#ifdef HAVE_PROGRESS
+       "-P command\n"
+       "\toutput progress indication to an external command\n"
+#endif /* HAVE_PROGRESS */
+#ifdef HAVE_CONVERTER
+       "-J command\n"
+       "\tconvert non-JPEG files to JPEG with an external command\n"
+#endif /* HAVE_CONVERTER */
        "--\tdisable processing further options\n",
        stderr);
     return 0;
@@ -544,7 +916,9 @@ main (int argc, char** argv)
 	goto Usage;
       case 'T':
 	if (argc <= 2) {
-	  (void) fputs ("-T: missing argument (use -h for help)\n", stderr);
+	missing:
+	  (void) fprintf (stderr, "-%c: missing argument (use -h for help)\n",
+			  *opts);
 	  return 1;
 	}
 	else {
@@ -563,10 +937,8 @@ main (int argc, char** argv)
 	}
 	break;
       case 'H':
-	if (argc <= 2) {
-	  (void) fputs ("-H: missing argument (use -h for help)\n", stderr);
-	  return 1;
-	}
+	if (argc <= 2)
+	  goto missing;
 	argc--;
 	if (!*(hopts.html = *++argv)) {
 	  (void) fputs ("-H: argument must be nonempty (use -h for help)\n",
@@ -581,10 +953,8 @@ main (int argc, char** argv)
 	}
 	break;
       case 'C':
-	if (argc <= 2) {
-	  (void) fputs ("-C: missing argument (use -h for help)\n", stderr);
-	  return 1;
-	}
+	if (argc <= 2)
+	  goto missing;
 	argc--;
 	if (!*(hopts.css = *++argv)) {
 	  (void) fputs ("-C: argument must be nonempty (use -h for help)\n",
@@ -593,27 +963,21 @@ main (int argc, char** argv)
 	}
 	break;
       case 'c':
-	if (argc <= 2) {
-	  (void) fputs ("-c: missing argument (use -h for help)\n", stderr);
-	  return 1;
-	}
+	if (argc <= 2)
+	  goto missing;
 	argc--;
 	hopts.charset = *++argv; /* no sanity check */
 	break;
       case 'i':
-	if (argc <= 2) {
-	  (void) fputs ("-i: missing argument (use -h for help)\n", stderr);
-	  return 1;
-	}
+	if (argc <= 2)
+	  goto missing;
 	argc--;
 	if (!*(hopts.icon = *++argv))
 	  hopts.icon = 0;
 	break;
       case 'I':
-	if (argc <= 2) {
-	  (void) fputs ("-I: missing argument (use -h for help)\n", stderr);
-	  return 1;
-	}
+	if (argc <= 2)
+	  goto missing;
 	argc--;
 	if (!*(hopts.msico = *++argv))
 	  hopts.msico = 0;
@@ -628,10 +992,8 @@ main (int argc, char** argv)
 	hopts.notim = 1;
 	break;
       case 'q':
-	if (argc <= 2) {
-	  (void) fputs ("-q: missing argument (use -h for help)\n", stderr);
-	  return 1;
-	}
+	if (argc <= 2)
+	  goto missing;
 	else {
 	  char* endp;
 	  unsigned long q = strtoul (*++argv, &endp, 0);
@@ -642,6 +1004,57 @@ main (int argc, char** argv)
 	  }
 	  quality = q;
 	}
+	break;
+#ifdef HAVE_LINK
+      case 'l':
+      case 'L':
+	if (argc <= 2)
+	  goto missing;
+	argc--;
+	linkdir = *++argv;
+	linkdirlen = 0;
+	if (!*linkdir)
+	  linkdir = 0;
+	if (linkdir) {
+	  linkdirlen = strlen (linkdir);
+	  /* strip trailing slashes */
+	  while (linkdirlen && linkdir[linkdirlen - 1] == '/')
+	    linkdirlen--;
+	  linkdirlen++;
+	}
+	linkdir_sym = *opts == 'l';
+	break;
+#endif /* HAVE_LINK */
+#ifdef HAVE_PROGRESS
+      case 'P':
+	if (argc <= 2)
+	  goto missing;
+	if (progress)
+	  pclose (progress);
+	argc--;
+	fflush (stderr);
+	fflush (stdout);
+	if (*++argv) {
+	  progress = popen (*argv, "w");
+	  if (!progress) {
+	    fprintf (stderr, "-P %s", *argv);
+	    perror (": popen");
+	  }
+	}
+	else
+	  progress = 0;
+	break;
+#endif /* HAVE_PROGRESS */
+#ifdef HAVE_CONVERTER
+      case 'J':
+	if (argc <= 2)
+	  goto missing;
+	argc--;
+	converter = *++argv;
+	if (!*converter)
+	  converter = 0;
+	break;
+#endif /* HAVE_CONVERTER */
       }
     }
   }
@@ -660,6 +1073,10 @@ main (int argc, char** argv)
     unsigned num_slash;
     if (!*argv)
       continue; /* ignore empty directory names */
+#ifdef HAVE_LINK
+    if (linkdir && make_directory (*argv))
+      return 2;
+#endif /* HAVE_LINK */
     if (dump_html && !remove_thumbdirs && create_css (*argv))
       return 2;
     else {
@@ -675,13 +1092,38 @@ main (int argc, char** argv)
 	num_slash--;
       c[1] = 0;
     }
-    inittree (&dirs, *argv);
-    if (readtree (&dirs, prune_dirs) ||
-	(freefileinfo (&dirs.n_others, &dirs.others),
-	 dumptree (&dirs, num_slash, 0, 0, 0))) {
+
+#ifdef HAVE_LINK
+    if (linkdir) {
+      char* path;
+      size_t namelen = strlen (*argv);
+
+      if (!(path = malloc (linkdirlen + namelen + 1))) {
+	(void) fputs (*argv, stderr);
+	(void) fputs (": out of memory\n", stderr);
+	return 2;
+      }
+
+      memcpy (path, linkdir, linkdirlen - 1);
+      path[linkdirlen - 1] = '/';
+      memcpy (path + linkdirlen, *argv, namelen + 1);
+
+      inittree (&dirs, path);
+      free (path);
+    }
+    else
+#endif /* HAVE_LINK */
+      inittree (&dirs, *argv);
+    if (readtree (&dirs, prune_dirs, identify_images)) {
+    tree_error:
       freetree (&dirs);
       return 2;
     }
+#ifdef HAVE_PROGRESS
+    progresstotal += dirs.n_files + dirs.n_files_dirs;
+#endif /* HAVE_PROGRESS */
+    if (dumptree (&dirs, num_slash, 0, 0, 0))
+      goto tree_error;
     freetree (&dirs);
   }
 
