@@ -3,7 +3,7 @@
  * @author Marko Mkel <marko.makela@iki.fi>
  */
 
-/* Copyright  2003 Marko Mkel.
+/* Copyright  2003,2006 Marko Mkel.
 
    This file is part of PHOTOMOLO, a program for generating
    thumbnail images and HTML files for browsing digital photographs.
@@ -69,11 +69,13 @@ cmp_files (const void* a,
 /** Read a directory tree
  * @param tree	the destination tree (with the path name filled in)
  * @param dir_p	callback for pruning directories by name
+ * @param img_p	callback for pruning non-image files by name
  * @return	zero on success, nonzero on failure
  */
 int
 readtree (struct dirtree* tree,
-	  int (*dir_p) (const char*))
+	  int (*dir_p) (const char*),
+	  enum file_type (*img_p) (const char*))
 {
   /** directory handle */
   DIR* dir;
@@ -140,37 +142,8 @@ readtree (struct dirtree* tree,
       inittree (&tree->dirs[tree->n_dirs++], pathname);
     }
     else if (S_ISREG (statbuf.st_mode)) {
-      /* regular file; is it a readable image file? */
-      FILE* f = fopen (pathname, "rb");
-      if (!f) {
-	(void) fputs (pathname, stderr), (void) fflush (stderr);
-	perror (": open (reading)");
-      other:
-	/* non-image file */
-	if (!(tree->n_others & (tree->n_others + 1))) {
-	  /* double the size allocated for the file name buffer */
-	  struct fileinfo* others =
-	    realloc (tree->others,
-		     (tree->n_others + 1) * (2 * sizeof *others));
-	  if (!others)
-	    goto memory;
-	  tree->others = others;
-	}
-	if (!(tree->others[tree->n_others].name = strdup (dirent->d_name)))
-	  goto memory;
-	tree->others[tree->n_others].mtime = statbuf.st_mtime;
-	tree->n_others++;
-	/* non-image files do not affect tree->mtime */
-	continue;
-      }
-      else {
-	int is_jpeg =
-	  getc (f) == 0xff &&
-	  getc (f) == 0xd8;
-	(void) fclose (f);
-	if (!is_jpeg)
-	  goto other;
-      }
+      enum file_type kind = (*img_p) (pathname);
+      struct fileinfo* file;
       if (!(tree->n_files & (tree->n_files + 1))) {
 	/* double the size allocated for the image file name buffer */
 	struct fileinfo* files =
@@ -179,12 +152,23 @@ readtree (struct dirtree* tree,
 	  goto memory;
 	tree->files = files;
       }
-      if (!(tree->files[tree->n_files].name = strdup (dirent->d_name)))
+      file = &tree->files[tree->n_files];
+      if (!(file->name = strdup (dirent->d_name)))
 	goto memory;
-      tree->files[tree->n_files].mtime = statbuf.st_mtime;
+      file->mtime = statbuf.st_mtime;
+      file->kind = kind;
+      file->converted = 0;
       tree->n_files++;
-      if (tree->mtime < statbuf.st_mtime)
-	tree->mtime = statbuf.st_mtime;
+      switch (kind) {
+      case FILE_JPEG:
+	tree->n_jpeg++;
+      case FILE_IMAGE:
+	if (tree->mtime < statbuf.st_mtime)
+	  tree->mtime = statbuf.st_mtime;
+	break;
+      case FILE_OTHER:
+	break;
+      }
     }
   }
 
@@ -193,18 +177,20 @@ readtree (struct dirtree* tree,
 
   /* process the subdirectories */
   for (i = 0; i < tree->n_dirs; i++) {
-    int status = readtree (tree->dirs + i, dir_p);
+    int status = readtree (tree->dirs + i, dir_p, img_p);
     if (status) {
       freetree (tree);
       return status;
     }
     if (tree->mtime < tree->dirs[i].mtime)
       tree->mtime = tree->dirs[i].mtime;
+    tree->n_files_dirs +=
+      tree->dirs[i].n_files_dirs + tree->dirs[i].n_files;
+    tree->n_jpeg_dirs +=
+      tree->dirs[i].n_jpeg_dirs + tree->dirs[i].n_jpeg;
   }
   /* sort the image files */
   qsort (tree->files, tree->n_files, sizeof *tree->files, cmp_files);
-  /* sort other files */
-  qsort (tree->others, tree->n_others, sizeof *tree->others, cmp_files);
   /* sort the directories */
   qsort (tree->dirs, tree->n_dirs, sizeof *tree->dirs, cmp_files);
 
@@ -215,13 +201,17 @@ readtree (struct dirtree* tree,
  * @param num	pointer to number of entries
  * @param info	pointer to the array
  */
-void
+static void
 freefileinfo (unsigned* num,
 	      struct fileinfo** info)
 {
   unsigned i = *num;
-  while (i--)
-    free ((*info)[i].name);
+  struct fileinfo* fi = *info;
+  while (i--) {
+    free (fi->name);
+    free (fi->converted);
+    fi++;
+  }
   free (*info);
   *num = 0, *info = 0;
 }
@@ -235,7 +225,6 @@ freetree (struct dirtree* tree)
   register unsigned i;
   free (tree->name);
   freefileinfo (&tree->n_files, &tree->files);
-  freefileinfo (&tree->n_others, &tree->others);
   for (i = tree->n_dirs; i--; )
     freetree (tree->dirs + i);
   free (tree->dirs);
@@ -365,3 +354,26 @@ remove_directory (const char* path)
 
   return 0;
 }
+
+#ifdef HAVE_LINK
+/** Create a hard or soft (symbolic) link
+ * @param dname	the path name of the link destination
+ * @param name	the path name of the link source
+ * @param sym	nonzero=create symbolic link; zero=hard link
+ * @return	zero on success, nonzero on failure
+ */
+int
+create_link (const char* dname, const char* name, int sym)
+{
+  int st;
+  if (sym)
+    st = symlink (dname, name);
+  else
+    st = link (dname, name);
+  if (!st || errno == EEXIST)
+    return 0;
+  (void) fputs (name, stderr), (void) fflush (stderr);
+  perror (sym ? ": symlink" : ": link");
+  return 1;
+}
+#endif /* HAVE_LINK */
